
class MyTable : public Fl_Table {

	int kind;

	void DrawHeader(const char *s, int X, int Y, int W, int H) {
		fl_push_clip(X,Y,W,H);
		fl_draw_box(FL_THIN_UP_BOX, X,Y,W,H, row_header_color());
		fl_color(FL_BLACK);
		fl_draw(s, X,Y,W,H, FL_ALIGN_CENTER);
		fl_pop_clip();
	}

	void DrawData(const char *s, int X, int Y, int W, int H, Fl_Color bg) {
		fl_push_clip(X,Y,W,H);
		fl_color(bg); fl_rectf(X,Y,W,H);
		fl_color(FL_GRAY0); fl_draw(s, X,Y,W,H, FL_ALIGN_CENTER);
		fl_color(color()); fl_rect(X,Y,W,H);
		fl_pop_clip();
	} 

	void draw_cell(TableContext context, int ROW=0, int COL=0, int X=0, int Y=0, int W=0, int H=0) {
		static char s[40];
		int adr;
		Fl_Color color=FL_WHITE;
		switch ( context ) {
			case CONTEXT_STARTPAGE:
			fl_font(FL_HELVETICA, 11);
			return; 
			case CONTEXT_COL_HEADER:
			sprintf(s,"%X",COL);
			DrawHeader(s,X,Y,W,H);
			return; 
			case CONTEXT_ROW_HEADER:
			if(kind==DT_FLASH) sprintf(s,"%04X",ROW<<3);
			else sprintf(s,"%04X",ROW<<4);
			DrawHeader(s,X,Y,W,H);
			return; 
			case CONTEXT_CELL:
			switch(kind){
				case DT_RAM:
				adr=(ROW<<4)+COL;
				if(adr==last_modified_ram) color=FL_RED;
				sprintf(s,"%02X",ram[adr]);
				break;
				case DT_EEPROM:
				adr=(ROW<<4)+COL;
				if(adr==last_modified_eeprom) color=FL_RED;
				sprintf(s,"%02X",eeprom[adr]);
				break;
				case DT_FLASH:
				adr=(ROW<<3)+COL;
				if(adr==last_modified_flash) color=FL_RED;
				sprintf(s,"%04X",flash[adr].opcode);
				break;
				default:
				strcpy(s,"??");
				break;
			}
			DrawData(s,X,Y,W,H,color);
			return;
			default:
			return;
		}
	}

static void event_callback(Fl_Widget* wi, void* par)
  {
	MyTable *m;
	Fl_Table* t;
	const char *cp;
	int i,adr;
	t=(Fl_Table*) wi;
	m=(MyTable*) wi;
	TableContext k=t->callback_context();
	if(k==CONTEXT_CELL){
		if(m->kind==DT_FLASH) adr=(t->callback_row()<<3)+t->callback_col();
		else adr=(t->callback_row()<<4)+t->callback_col();
		switch(m->kind){

			case DT_RAM:
			cp=fl_input("RAM at 0x%04X","",adr);
			if(cp){
				sscanf(cp,"%x",&i);
				ram[adr]=i;
				last_modified_ram=adr;
				refresh();
			}
			break;

			case DT_EEPROM:
			cp=fl_input("EEPROM at 0x%04X","",adr);
			if(cp){
				sscanf(cp,"%x",&i);
				eeprom[adr]=i;
				last_modified_eeprom=adr;
				refresh();
			}
			break;

			case DT_FLASH:
			cp=fl_input("FLASH at 0x%04X","",adr);
			if(cp){
				sscanf(cp,"%x",&i);
				flash[adr].opcode=i;
				last_modified_flash=adr;
				refresh();
			}
			break;

			default:
			break;
		}
	}
  }

public:
  MyTable(int X, int Y, int W, int H, int kind=0,int col_width=20) : Fl_Table(X,Y,W,H) {
	callback(&event_callback, (void*)this);
	this->kind=kind;
    // Rows
    Fl_Table::rows(MAX_ROWS);
    row_header(1);
    row_height_all(16);
    row_resize(0);
    // Cols
    if(kind==DT_FLASH) cols(8);
    else cols(16);
    col_header(1);
    col_width_all(col_width);
    row_header_width(35);
    col_resize(1);
    end();
  }
  ~MyTable() { }
};

MyTable *ram_tab,*eeprom_tab,*flash_tab;
