#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <Fl/Fl.H>
#include <Fl/Fl_Double_Window.H>
#include <Fl/Fl_Input.H>
#include <FL/fl_ask.H>

#include <stdint.h>

#define PORT_DEFS		1
#define MEMORY_LIMITS	2

int section=-1;

typedef struct portrec{
	int port,bit;
	char name[20],longnam[200];
}portrec;

void actPortName(char *name, int val){
	int c=0;
	strcpy(io_names[val],name);
	while(c<MAX_PORTS){
		if(!strcmp(regNames[c],name)){
			ports[c]=val;
			return;
		}
		c++;
	}
}

void actBitName(char *name, int val){
	int c=0;
	while(c<MAX_BITS){
		if(!strcmp(bitNames[c],name)){
			bits[c]=val;
			return;
		}
		c++;
	}
}

void procLine(char *st){
	char *cp, *cp1;
	Fl_Tree_Item *ti;
	static int po,bi;
	portrec *pr;
	if(cp=strchr(st,'\r')) *cp=0;
	if(cp=strchr(st,'\n')) *cp=0;
	st[60]=0;
	//printf("%s\n",st);

	if(!strcmp(st,"[PortDefs]")){
		section=PORT_DEFS;
		return;
	}

	if(!strcmp(st,"[MemoryLimits]")){
		section=MEMORY_LIMITS;
		return;
	}

	if(section==-1){
		if(!strncmp(st,"Device=",7)){
			strcpy(devicename,st+7);
		}
	}

	if(section==PORT_DEFS){

		if(!strncmp(st,"Group ",6)){
			gr=tree->add(st+6);
			if(!gr){
				printf("Error creating group %s\n",st+6);
				return;
			}
			gr->close();
			gr->labelfont(FL_COURIER);
			return;
		}

		if(!strncmp(st," Port ",6)){
			cp=strchr(st+6,'=');
			if(!cp) return;
			*cp=0;
			cp++;
			sscanf(cp,"%x",&po);
			cp=strchr(cp,';');
			if(!cp) cp=st+6;
			else cp++;
			pr=(portrec*)calloc(sizeof(portrec),1);
			strcpy(pr->longnam,cp);
			strcpy(pr->name,st+6);
			pr->port=po;
			pr->bit=-1;
			if(!gr) return;
			reg=gr->add(Fl_Tree_Prefs(),st+6);
			if(!reg){
				printf("Error creating port %s\n",st+6);
				return;
			}
			reg->labelfont(FL_COURIER);
			reg->user_data((void*)pr);
			reg->close();
			actPortName(st+6,po);
			return;
		}

		if(!strncmp(st,"  Bit ",6)){
			cp=strchr(st+6,'=');
			if(!cp) return;
			*cp=0;
			cp++;
			bi=atoi(cp);
			cp=strchr(cp,';');
			if(!cp) return;
			cp++;
			pr=(portrec*)calloc(sizeof(portrec),1);
			strcpy(pr->longnam,cp);
			strcpy(pr->name,st+6);
			pr->port=po;
			pr->bit=bi;
			if(!reg) return;
			ti=reg->add(Fl_Tree_Prefs(),st+6);
			if(!ti){
				printf("Error creating bit %s\n",st+6);
				return;
			}
			ti->user_data((void*)pr);
			ti->labelfont(FL_COURIER);
			return;
		}
	}

}

void refreshTree(Fl_Tree_Item *ti){
	int n,c;
	char st[2048];
	portrec *pr;
	Fl_Tree_Item *ch;
	pr=(portrec*)ti->user_data();
	if(pr){
		if(pr->bit<0){
		sprintf(st,"%02x %-12s= %02x ;%s",pr->port,pr->name,ram[pr->port],pr->longnam);
		}
		else{
		sprintf(st,"%d %-12s= %d ;%s",pr->bit,pr->name,
			((ram[pr->port])>>pr->bit)&1,pr->longnam);
		}
		ti->label(st);
	}
	n=ti->children();
	if(!n) return;
	for(c=0;c<n;c++){
		ch=ti->child(c);
		refreshTree(ch);
	}
}

void cleanTree(Fl_Tree_Item *ti){
	int n,c;
	char st[2048];
	portrec *pr;
	Fl_Tree_Item *ch;
	pr=(portrec*)ti->user_data();
	if(pr) free(pr);
	n=ti->children();
	if(!n) return;
	for(c=0;c<n;c++){
		ch=ti->child(c);
		cleanTree(ch);
	}
}

void readDefFile(){
	FILE *f;
	int c;
	char st[2048];
	cleanTree(tree->root());
	tree->clear();
	section=-1;
	for(c=0;c<256;c++){
		sprintf(st,"0x%02x",c);
		strcpy(io_names[c],st);
	}
	f=fopen(definitions,"r");
	if(!f) return;
	while(fgets(st,2040,f)) procLine(st);
	fclose(f);
	refresh();
	refreshTitle();
}

void tree_callback(Fl_Widget *w){
	portrec *pr;
	const char *cp;
	int i;
	Fl_Tree *tree = (Fl_Tree*)w;
    Fl_Tree_Item *item = tree->callback_item();
	pr=(portrec*) item->user_data();
	if(!pr) return;
	Fl_Tree_Reason reason;
	reason=tree->callback_reason();
	if(reason==1){
		if(pr->bit<0){
			cp=fl_input("%s at 0x%04X","",pr->name,pr->port);
			if(cp){
				sscanf(cp,"%x",&i);
				ram[pr->port]=i;
			}
		}
		else{
			ram[pr->port]^=(1<<pr->bit);
		}
		refresh();
	}
}
